"""
Connection to power supply unit

"""
import time

from plx_gpib_ethernet import PrologixGPIBEthernet

from . import osci
from . import commands as cmd
from . import logging

bar_available = False

try:
    import pyprind
    bar_available = True
except ImportError:
    pass
    #logger.warning("No pyprind available")

setget = osci.setget
KCmd = cmd.KeysightE3631APowerSupplyCommands

q = cmd.query

class KeysightE3631APowerSupply(object):
    """
    A low volgage power supply with two channels, +6V and +- 25V manufactured
    by Keysight. The power supply does not have an ethernet port, so the
    connection is done via GPIB and a prologix GPIB Ethernet connector
    """
    output = setget(KCmd.OUTPUT)


    def __init__(self, ip="10.25.124.252", gpib_address=5, loglevel=20):
        """
        Connect to the power supply via Prologix GPIB connector

        Keyword Args:
            ip (str): IP adress of the Prologix GPIB connector
            gpib_address (int): The GPIB address of the power supply
                                connected to the Prologix connector
        """
        gpib = PrologixGPIBEthernet(ip)
        gpib.connect()
        gpib.select(gpib_address)
        self.logger = logging.get_logger(loglevel)
        self.instrument = gpib
        self.P6 = KCmd.P6
        self.P25 = KCmd.P25
        self.N25 = KCmd.N25

    def __del__(self):
        self.instrument.close()

    def _query(self, command):
        """
        Send a command

        Args:
            command (str):

        Returns:
            str
        """
        self.logger.debug("Querying {}".format(command))
        return self.instrument.query(command)

    def _set(self, command):
        """
        Send a command bur return no response

        Args:
            command (str): command to be send to the scope

        Returns:
            None
        """

        #FIXME: Make AbstractInstrment class and inherit
        self.logger.debug("Sending {}".format(command))
        self.instrument.write(command)

    def ping(self):
        """
        Check the connection

        Returns:
            str
        """

        return self.instrument.query(cmd.WHOAMI)

    @property
    def error_state(self):
        """
        Read out the error register of the power supply

        Returns:
            str
        """

        error =  self.instrument.query(KCmd.ERROR_STATEQ)
        error = error.split(",")
        err_no = float(error[0])
        return err_no, "".join(error[1:])

    def select_channel(self, channel):
        """
        Select either the +6, +25 or -25V channel

        Args:
            channel (str or int):

        Returns:
            None
        """
        channel_dict = {1: self.P6, 2: self.P25, 3: self.N25}
        if isinstance(channel, int):
            assert channel in [1, 2, 3], "Channel has to be either 0:+6V, 1:+25V or 2: -25V"
            channel = channel_dict[channel]
        elif isinstance(channel, str):
            assert channel in (KCmd.P6, KCmd.P25, KCmd.N25),\
                "Channel has to be in {}".format(KCmd.P6, KCmd.P25, KCmd.N25)
        else:
            raise ValueError("Channel has to be either str or int")

        self._set(KCmd.CHANNEL + " {}".format(channel))
        return channel

    def set_voltage(self, channel, voltage):
        """
        Set the supplied voltage of a channel to the desired value

        Args:
            channel (str or int):
            voltage (float):

        Returns:
            None
        """
        channel = self.select_channel(channel)
        if (channel == self.P6) and (voltage > 6.18):
            raise ValueError("6V Channel does not support {}".format(voltage))
        self._set(KCmd.VOLT + " {}".format(voltage))


    def off(self):
        """
        Cut the power on all channels

        Returns:
            None
        """
        self.logger.info("Disabling power...")
        self.output = KCmd.OFF

    def on(self):
        """
        Enable power on all channels

        Returns:
            None
        """
        self.logger.warning("Enabling power!!")

        # give the user time to hit Ctrl-C to abort
        safety_wait = 5
        if bar_available:
            bar = pyprind.ProgBar(safety_wait, track_time=True, title='Powering on...')

        for i in range(safety_wait):
            time.sleep(1)
            if bar_available:
                bar.update()
            else:
                print ("Powering on in {}".format(safety_wait - i))
        self.output = KCmd.ON

    def measure_current(self, channel):
        """
        Measure current on givven channel

        Args:
            channel (str):

        Returns:
            float
        """
        command = q(KCmd.MEASURE + ":" + KCmd.CURRENT + ":" + KCmd.DC)
        command += " {}".format(channel)
        return float(self._query(command))





"""
Provide routines for fitting charge histograms
"""

import sys
import pylab as p
import numpy as np
from scipy.misc import factorial

from functools import reduce
from copy import deepcopy as copy
from collections import namedtuple

import scipy.optimize as optimize
import seaborn.apionly as sb
import dashi as d

from . import tools
from . import plotting as plt


from scipy.constants import elementary_charge as ELEMENTARY_CHARGE
d.visual()

try:
    from iminuit import Minuit
except ImportError:
    print ("WARNING, can not load iminuit")

# default color palette
PALETTE = sb.color_palette("dark")
p.style.use("pyoscipresent")


def get_n_hit(charges, nbins):
    """
    Identify how many events are in the pedestal

    Args:
        charges (np.ndarray): The measured charges
        nbins (int): number of bins to use

    Returns:
        tuple (n_hit, n_all)
    """
    one_gauss = lambda x, n, y, z: n * gauss(x, y, z, 1)
    ped_mod = Model(one_gauss, (1000, -.1, 1))
    ped_mod.fit_to_data(charges, nbins, normalize=False, silent=False)
    n_hit = abs(ped_mod.data.bincontent - ped_mod.prediction(ped_mod.xs)).sum()
    n_pedestal = ped_mod.data.stats.nentries - n_hit
    n_all = ped_mod.data.stats.nentries
    return n_hit, n_all


def calculate_mu(charges, nbins):
    """
    Calculate mu out of
    P(hit) = (N_hit/N_all) = exp(QExCExLY)
    where P is the probability for a hit, QE is quantum efficiency, CE is
    """
    n_hit, n_all = get_n_hit(charges, nbins)
    n_pedestal = n_all - n_hit
    #mu = -1 * np.log(n_pedestal / n_all)
    mu = -1 * np.log(1 - (n_hit/n_all))
    return mu


def reject_outliers(data, m=2):
    """
    A simple way to remove extreme outliers from data

    Args:
        data (np.ndarray): data with outliers
        m (int): number of standard deviations outside the
                 data should be discarded

    Returns:
        np.ndarray
    """

    return data[abs(data - np.mean(data)) < m * np.std(data)]


def calculate_gain(mu_ped, mu_spe, prefactor=1e-12):
    """
    Calculate the pmt gain from the charge distribution

    Args:
        mu_ped (float): the mean of the gaussian fitting the pedestal
        mu_spe (float): the mean of the gaussian fitting the spe response

    Keyword Args:
        prefactor (float): unit of charge (default pico coulomb)

    Returns:
        float
    """
    charge =  abs(mu_spe) - abs(mu_ped)
    charge *= prefactor
    return charge/ELEMENTARY_CHARGE


def chi2_exponential_part(xs, data, pred, mu_ped, mu_ser):
    mask = np.logical_and(xs >= mu_ped, xs <= mu_ser)
    data = data[mask]
    pred = pred[mask]
    xs   = xs[mask]
    chi2_ndf = calculate_chi_square(data, pred)/len(xs)
    return chi2_ndf


def calculate_peak_to_valley_ratio(bestfitmodel, mu_ped, mu_spe, control_plot=False):
    """
    Calculate the peak to valley ratio
    Args:
        bestfitmodel (fit.Model): A fitted model to charge response data
        mu_ped (float): The x value of the fitted pedestal
        mu_spe (flota): The x value of the fitted spe peak
   
    Keyword Args:
        control_plot (bool): Show control plot to see if correct values are found
   
    """
    

    tmpdata = bestfitmodel.prediction(bestfitmodel.xs)
    valley = min(tmpdata[np.logical_and(bestfitmodel.xs > mu_ped,\
                                    bestfitmodel.xs < mu_spe)])
    valley_x = bestfitmodel.xs[tmpdata == valley]

    peak = max(tmpdata[bestfitmodel.xs > valley_x])
    peak_x = bestfitmodel.xs[tmpdata == peak]
    peak_v_ratio = (peak/valley)
    
    if control_plot:
        fig = p.figure()
        ax = fig.gca()
        ax.plot(bestfitmodel.xs,tmpdata)
        print (valley)
        print (valley_x)

        ax.scatter(valley_x,valley,marker="o")
        ax.scatter(peak_x, peak, marker="o")
        ax.set_ylim(ymin=1e-4)
        ax.set_yscale("log")
        ax.grid(1)
        sb.despine()

    return peak_v_ratio


def calculate_chi_square(data, model_data):
    """
    Very simple estimator for goodness-of-fit. Use with care.
    Non normalized bin counts are required.

    Args:
        data (np.ndarray): observed data (bincounts)
        model_data (np.ndarray): model predictions for each bin

    Returns:
        np.ndarray
    """

    chi = ((data - model_data)**2/data)
    return chi[np.isfinite(chi)].sum()


def poisson(lmbda, k):
    """
    Poisson distribution

    Args:
        lmbda (int): expected number of occurences
        k (int): measured number of occurences

    Returns:
        np.ndarrya
    """

    return np.power(lmbda, k) * np.exp(-1 * lmbda) / factorial(k)


def gauss(x, mu, sigma, n):
    """
    Returns a normed gaussian.

    Args:
        x (np.ndarray): x values
        mu (float): Gauss mu
        sigma (float): Gauss sigma
        n:

    Returns:

    """

    def _gaussnorm(sigma, n):
        return 1 / (sigma * np.sqrt(2 * n * np.pi))

    return _gaussnorm(sigma, n) * np.exp(-np.power((x - (n * mu)), 2) / (2 * n * (sigma ** 2)))


def calculate_sigma_from_amp(amp):
    """
    Get the sigma for the gauss from its peak value.
    Gauss is normed

    Args:
        amp (float):

    Returns:
        float
    """
    return 1/(np.sqrt(2*np.pi)*amp)


def exponential(x, N, n_noise):
    """
    An exponential model (for noise?)

    Args:
        x:
        N:
        n_noise:

    Returns:
        np.ndarray
    """
    return N * np.exp(-n_noise * x)


class Model(object):
    """
    Model holds theoretical prediction for
    data. Can be used to fit data.

    """

    def __init__(self, func, startparams=None, norm=1):
        """
        Initialize a new model


        Args:
            func: the function to predict the data
            startparams (tuple): A set of startparameters.
            norm: multiply the result of func when evaluated with norm
        """

        # if no startparams are given, construct 
        # and initialize with 0es.
        # FIXME: better estimate?
        if startparams is None:
            startparams = [0]*func.__code__.co_argcount

        def normed_func(*args, **kwargs):
            return norm*func(*args, **kwargs)

        self._callbacks = [normed_func]
        self.startparams = [*startparams]
        self.n_params = [len(startparams)]
        self.best_fit_params = [*startparams]
        self.coupling_variable = []
        self.all_coupled = False
        self.data = None
        self.xs = None
        self.chi2_ndf = None
        self.chi2_ndf_components = []
        self.norm = None
        self.prediction = lambda xs: reduce(lambda x, y: x + y,\
                                  [f(xs) for f in self.components])
        self.first_guess = None

    def add_first_guess(self, func):
        """
        Use func to estimate better startparameters

        Args:
            func: Has to yield a set of startparameters

        Returns:

        """

        assert self.all_coupled, "Does not work yet if not all variables are coupled"
        assert func.__code__.co_argcount == len(self.startparams), "first guess algorithm must yield startparams!"
        self.first_guess = func

    def eval_first_guess(self, data):
        self.startparams = self.first_guess(data)

    def couple_models(self, coupling_variable):
        """
        Couple the models by a variable, which means use the variable not
        independently in all model components, but fit it only once.
        E.g. if there are 3 models with parameters p1, p2, k each and they
        are coupled by k, parameters p11, p21, p12, p22, and k will be fitted
        instead of p11, p12, k1, p21, p22, k2.

        Args:
            coupling_variable: variable number of the number in startparams

        Returns:
            None
        """
        assert len(np.unique(self.n_params)) == 1,\
            "Models have different numbers of parameters,difficult to couple!"

        self.coupling_variable.append(coupling_variable)

    def couple_all_models(self):
        """
        Use the first models startparams for
        the combined model

        Returns:
            None
        """
        self.all_coupled = True
        # if self.all_coupled:
        startparams = self.startparams[0:self.n_params[0]]

    def __add__(self, other):
        self._callbacks = self._callbacks + other._callbacks
        self.startparams = self.startparams + other.startparams
        self.n_params = self.n_params + other.n_params
        self.best_fit_params = self.startparams
        return self

    @property
    def components(self):
        lastslice = 0
        thecomponents = []
        for i, cmp in enumerate(self._callbacks):
            thisslice = slice(lastslice,self.n_params[i] + lastslice)
            tmpcmp = copy(cmp)
            #thecomponents.append(lambda xs: tmpcmp(xs, ),*self.best_fit_params[thisslice])
            lastslice += self.n_params[i]
            best_fit = self.best_fit_params[thisslice]
            if self.all_coupled:
                best_fit = self.best_fit_params[0:self.n_params[0]]
            yield lambda xs: tmpcmp(xs, *best_fit)
        #return thecomponents

    def __call__(self, xs, *params):
        """
        Give the model prediction

        Args:
            xs (np.ndaarray): the values the model should be evaluated on

        Returns:
            np.ndarray
        """
        thecomponents = []
        firstparams = params[0:self.n_params[0]]
        first = self._callbacks[0](xs, *firstparams)

        lastslice = self.n_params[0]
        for i, cmp in enumerate(self._callbacks[1:]):
            thisslice = slice(lastslice, self.n_params[1:][i] + lastslice)
            # tmpcmp = copy(cmp)
            theparams = list(params[thisslice])
            if self.coupling_variable:
                for k in self.coupling_variable:
                    theparams[k] = firstparams[k]
            elif self.all_coupled:
                theparams = firstparams
            # thecomponents.append(lambda xs: tmpcmp(xs, *params[thisslice]))
            lastslice += self.n_params[1:][i]
            first += cmp(xs, *theparams)
        return first

    def add_data(self, data, nbins, subtract = None):
        """
        Add some data to the model, in preparation for the fit


        Args:
            data:
            nbins:
            subtract:

        Returns:

        """
        bins = np.linspace(min(data), max(data), nbins)
        self.data = d.factory.hist1d(data, bins).normalized(density=True)
        self.xs = self.data.bincenters

    def fit_to_data(self, data, nbins, silent=False, subtract=None,\
                    normalize=True, **kwargs):
        """
        Apply this model to data

        Args:
            data (np.ndarray): the data, unbinned
            nbins (int): number of bins to put the data in
            silent (bool): silence output
            subtract (func): subtract this from the data before fitting
            normalize (bool): normalize data before fitting
            **kwargs: will be passed on to scipy.optimize.curvefit

        Returns:
            None
        """
        def model(xs, *params):
            thecomponents = []
            firstparams = params[0:self.n_params[0]]
            first = self._callbacks[0](xs, *firstparams)

            lastslice = self.n_params[0]
            for i, cmp in enumerate(self._callbacks[1:]):
                thisslice = slice(lastslice, self.n_params[1:][i] + lastslice)
                #tmpcmp = copy(cmp)
                theparams = list(params[thisslice])
                if self.coupling_variable:
                    for k in self.coupling_variable:
                        theparams[k] = firstparams[k]
                elif self.all_coupled:
                    theparams = firstparams
                #thecomponents.append(lambda xs: tmpcmp(xs, *params[thisslice]))
                lastslice += self.n_params[1:][i]
                first += cmp(xs, *theparams)
            return first

        startparams = self.startparams
        #if self.all_coupled:
        #    startparams = self.startparams[0:self.n_params[0]]
        bins = np.linspace(min(data), max(data), nbins)
        self.data = d.factory.hist1d(data, bins)
        if normalize:
            self.data = self.data.normalized(density=True)
        self.xs = self.data.bincenters

        h = d.factory.hist1d(data, bins)
        h_norm = h.normalized(density=True)
        xs = h_norm.bincenters
        data = self.data.bincontent
        if subtract is not None:
            data -= subtract(self.xs)

        if not silent: print("Using start params...", startparams)

        fitkwargs = {"maxfev": 1000000, "xtol": 1e-10, "ftol": 1e-10}
        if "bounds" in kwargs:
            fitkwargs.pop("maxfev")
            # this is a matplotlib quirk
            fitkwargs["max_nfev"] = 1000000
        fitkwargs.update(kwargs)
        parameters, covariance_matrix = optimize.curve_fit(model, self.xs,\
                                                           data, p0=startparams,\
                                                           # bounds=(np.array([0, 0, 0, 0, 0] + [0]*len(start_params[5:])),\
                                                           # np.array([np.inf, np.inf, np.inf, np.inf, np.inf] +\
                                                           # [np.inf]*len(start_params[5:]))),\
                                                           # max_nfev=100000)
                                                           # method="lm",\
                                                           **fitkwargs)

        if not silent: print("Fit yielded parameters", parameters)
        if not silent: print("{:4.2f} NANs in covariance matrix".format(len(covariance_matrix[np.isnan(covariance_matrix)])))
        if not silent: print("##########################################")

        # simple GOF
        norm = 1
        if normalize:
            norm = h.bincontent / h_norm.bincontent
            norm = norm[np.isfinite(norm)][0]

        self.norm = norm
        chi2 = (calculate_chi_square(h.bincontent, norm * model(h.bincenters, *parameters)))
        self.chi2_ndf = chi2/nbins

        # FIXME: new feature
        #for cmp in self.components:
        #    thischi2 = (calculate_chi_square(h.bincontent, norm * cmp(h.bincenters)))
        #    self.chi2_ndf_components.append(thischi2/nbins)

        if not silent: print("Obtained chi2 and chi2/ndf of {:4.2f} {:4.2f}".format(chi2, chi2 / nbins))
        self.best_fit_params = parameters
        return parameters
        #self.best_fit_params = fit_model(data, nbins, model, startparams, **kwargs)

    def clear(self):
        """
        Reset the model

        Returns:
            None
        """
        self.__init__(self._callbacks[0], self.startparams[:self.n_params])

    def plot_result(self, ymin=1000,xmax=8, ylabel="normed bincount",\
                    xlabel="Q [C]", fig=None,\
                    log=True,\
                    model_alpha=.3,\
                    add_parameter_text=((r"$\mu_{{SPE}}$& {:4.2e}\\",0),)):
        """
        Show the fit result

        Args:
            ymin (float): limit the yrange to ymin
            xmax (float): limit the xrange to xmax
            model_alpha (float): 0 <= x <= 1 the alpha value of the lineplot
                                for the model
            ylabel (str): label for yaxis
            log (bool): plot in log scale
            fig (pylab.figure): A figure instance
            add_parameter_text (tuple): Display a parameter in the table on the plot
                                        ((text, parameter_number), (text, parameter_number),...)
        Returns:
            pylab.figure
        """
        if fig is None:
            fig = p.figure()
        ax = fig.gca()
        self.data.scatter(color="k")
        ax.plot(self.xs, self.prediction(self.xs), color=PALETTE[2], alpha=model_alpha)
        for comp in self.components:
            ax.plot(self.xs, comp(self.xs), linestyle=":", lw=1, color="k")

        if log: ax.set_yscale("log")
        ax.set_ylim(ymin=ymin)
        ax.set_xlim(xmax=xmax)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        infotext = r"\begin{tabular}{ll}"
        infotext += r"$\chi^2/ndf$ & {:4.2f}\\".format(self.chi2_ndf)
        infotext += r"entries& {}\\".format(self.data.stats.nentries)
        if add_parameter_text is not None:
            for partext in add_parameter_text:
                infotext += partext[0].format(self.best_fit_params[partext[1]])
            #infotext += r"$\mu_{{SPE}}$& {:4.2e}\\".format(self.best_fit_params[mu_spe_is_par])
        infotext += r"\end{tabular}"
        ax.text(0.9, 0.9, infotext,
            horizontalalignment='center',
            verticalalignment='center',
            transform=ax.transAxes)
        sb.despine()
        return fig


def pedestal_fit(filename, nbins, fig=None):
    """
    Fit a pedestal to measured waveform data
    One shot function for
    * integrating the charges
    * making a histogram
    * fitting a simple gaussian to the pedestal
    * calculating mu
        P(hit) = (N_hit/N_all) = exp(QExCExLY)
        where P is the probability for a hit, QE is quantum efficiency,
        CE is the collection efficiency and
        LY the (unknown) light yield

    Args:
        filename (str): Name of the file with waveform data
        nbins (int): number of bins for the underlaying charge histogram

    """

    head, wf = tools.load_waveform(filename)
    charges = -1e12 * tools.integrate_wf(head, wf)
    plt.plot_waveform(head, tools.average_wf(wf))
    p.savefig(filename.replace(".npy", ".wf.pdf"))
    one_gauss = lambda x, n, y, z: n * fit.gauss(x, y, z, 1)
    ped_mod = fit.Model(one_gauss, (1000, -.1, 1))
    ped_mod.fit_to_data(charges, nbins, normalize=False)
    fig = ped_mod.plot_result(add_parameter_text=((r"$\mu_{{ped}}$& {:4.2e}\\", 1), \
                                                  (r"$\sigma_{{ped}}$& {:4.2e}\\", 2)), \
                              xlabel=r"$Q$ [pC]", ymin=1, xmax=8, model_alpha=.2, fig=fig, ylabel="events")

    ax = fig.gca()
    n_hit = abs(ped_mod.data.bincontent - ped_mod.prediction(ped_mod.xs)).sum()
    ax.grid(1)
    bins = np.linspace(min(charges), max(charges), nbins)
    data = d.factory.hist1d(charges, bins)
    n_pedestal = ped_mod.data.stats.nentries - n_hit

    mu = -1 * np.log(n_pedestal / ped_mod.data.stats.nentries)

    print("==============")
    print("All waveforms: {:4.2f}".format(ped_mod.data.stats.nentries))
    print("HIt waveforms: {:4.2f}".format(n_hit))
    print("NoHit waveforms: {:4.2f}".format(n_pedestal))
    print("mu = -ln(N_PED/N_TRIG) = {:4.2e}".format(mu))

    ax.fill_between(ped_mod.xs, 1e-4, ped_mod.prediction(ped_mod.xs),\
                    facecolor=PALETTE[2], alpha=.2)
    p.savefig(filename.replace(".npy", ".pdf"))
    # xs = self.data.bincenters
    return ped_mod

################################################

def fit_model(charges, model, startparams=None, \
              rej_outliers=False, nbins=200, \
              silent=False,\
              parameter_text=((r"$\mu_{{SPE}}$& {:4.2e}\\", 5),),
              use_minuit=False,\
              normalize=True,\
              **kwargs):
    """
    Standardazied fitting routine

    Args:
        charges (np.ndarray): Charges obtained in a measurement (no histogram)
        model (pyosci.fit.Model): A model to fit to the data
        startparams (tuple): initial parameters to model, or None for first guess

    Keyword Args:
        rej_outliers (bool): Remove extreme outliers from data
        nbins (int): Number of bins
        parameter_text (tuple): will be passed to model.plot_result
        use_miniuit (bool): use minuit to minimize startparams for best 
                            chi2
        normalize (bool): normalize data before fitting
        silent (bool): silence output
    Returns:
        tuple
    """
    if rej_outliers:
        charges = reject_outliers(charges)
    if use_minuit:

        from iminuit import Minuit

        # FIXME!! This is too ugly. Minuit wants named parameters ... >.<

        assert len(startparams) > 10; "Currently more than 10 paramters are not supported for minuit fitting!"
        assert model.all_coupled, "Minuit fitting can only be done for models with all parmaters coupled!"
        names = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k"]

        funcstring = "def do_min("
        for i,__ in enumerate(startparams):
            funcstring += names[i] + ","
        funcstring = funcstring[:-1] + "):\n"
        funcstring += "\tmodel.startparams = ("
        for i,__ in enumerate(startparams):
            funcstring += names[i] + ","
        funcstring = funcstring[:-1] + ")\n"
        funcstring += "\tmodel.fit_to_data(charges, nbins, silent=True, **kwargs)"
        funcstring += "\treturn model.chi2_ndf"


        #def do_min(a, b, c, d, e, f, g, h, i, j, k): #FIXME!!!
        #    model.startparams = (a, b, c, d, e, f, g, h, i, j, k)
        #    model.fit_to_data(charges, nbins, silent=True, **kwargs)
        #    return model.chi2_ndf
        exec(funcstring)
        bnd = kwargs["bounds"]
        if "bounds" in kwargs:
            min_kwargs = dict()
            for i,__ in enumerate(startparams):
                min_kwargs["limit_" + names[i]] =(bnd[0][i],bnd[1][i])
            m = Minuit(do_min, **min_kwargs)
            #m = Minuit(do_min, limit_a=(bnd[0][0],bnd[1][0]),
            #                   limit_b=(bnd[0][1],bnd[1][1]),
            #                   limit_c=(bnd[0][2],bnd[1][2]),
            #                   limit_d=(bnd[0][3],bnd[1][3]),
            #                   limit_e=(bnd[0][4],bnd[1][4]),
            #                   limit_f=(bnd[0][5],bnd[1][5]),
            #                   limit_g=(bnd[0][6],bnd[1][6]),
            #                   limit_h=(bnd[0][7],bnd[1][7]),
            #                   limit_i=(bnd[0][8],bnd[1][8]),
            #                   limit_j=(bnd[0][9],bnd[1][9]),
            #                   limit_k=(bnd[0][10],bnd[1][10]))
        else:



            m = Minuit(do_min)
        # hand over the startparams
        for key, value in zip(["a","b","c","d","e","f","g","h","i","j"], startparams):
            m.values[key] = value
        m.migrad()
    else:
        model.startparams = startparams
        model.fit_to_data(charges, nbins,normalize=normalize, silent=silent, **kwargs)

    # check for named tuple
    if hasattr(startparams, "_make"): # duck typing
        best_fit_params = startparams._make(model.best_fit_params)
    else:
        best_fit_params = model.best_fit_params
    print("Best fit parameters {}".format(best_fit_params))

    return model

############################################



if __name__ == "__main__":

    charge_factor = 1/ELEMENTARY_CHARGE
    conversion = charge_factor/1e10
    conversion = 1.
    charge_factor = 1e10
    print (charge_factor)
    print (conversion)
    nbins = 70
    charges = abs(np.load(sys.argv[1])) * charge_factor #1e10
    print(charges)
    bin_edges = np.linspace(min(charges), max(charges), nbins)
    bin_centers = bin_edges[:-1] + 0.5 * (bin_edges[1] - bin_edges[0])
    hist, edges = np.histogram(charges, bin_edges, normed=False)
    histvals, __ = np.histogram(charges, bin_edges, density=True, normed=True)

    # mu_p, sigma_p, mu, sigma , lmbda
    start_params = np.array([0, 10, 1, .5, 0.] + [1] * 10)  # ,1,1,1])
    # mu_p, sigma_p, mu, sigma , N, n_noise
    start_params_noise = np.array([10, 0, 1*conversion, .1, 2*conversion , .5*conversion , 1., 1*conversion ])
    #start_params_noise = np.array([1, 1, 1, 1, 1, 1, 1, 1])
    #params_uno = fit_model(bin_centers, histvals, charge_response_model, start_params)
    params_duo = fit_model(bin_centers, histvals, charge_response_with_noise, start_params_noise)
    #pred_uno = charge_response_model(bin_centers, *params_uno)
    pred_duo = charge_response_with_noise(bin_centers, *params_duo)

    #pedestal_uno, contribs_uno = charge_response_model(bin_centers, *params_uno, return_ped=True)

    # for i,name in zip(cpars,["mu_p","sigma_p", "mu", "sigma", "lambda"] + ["w"]*len(cpars[5:])):
    #    print (i,name)
    # mu_p, sigma_p, mu, sigma , lmbda

    print (params_duo[4])
    mu_peak = (1e-10)*params_duo[4]*50*(1e-12) # units from integration mV, ns, and 50 ohms impedance and the 1e-10 we had in the beginning

    print (ELEMENTARY_CHARGE)
    gain = mu_peak/ELEMENTARY_CHARGE
    print (mu_peak)
    print (gain)
    fig = p.figure()
    ax = fig.gca()
    # evaluate contributions
    for contrib in charge_response_with_noise(bin_centers, *params_duo, return_components=True):
        ax.plot(bin_centers, contrib, linestyle=":", lw=1, color="k")

    #ax.plot(bin_centers, pedestal_uno, linestyle=":", lw=1, color="k")
    #
    #for i, contrib in enumerate(contribs_uno):
    #    ax.plot(bin_centers, contrib, linestyle=":", lw=1, color="k")

    # ax.plot(bin_centers, pred_uno, lw=2, color=PALETTE[0], label="Multi PE model")
    ax.plot(bin_centers, pred_duo, lw=2, color=PALETTE[1], label="Noise model")
    h = d.factory.hist1d(charges,nbins)
    #h.bincontent = h.bincontent/h.bincontent.sum()
    #h.binerror = h.binerror/h.bincontent.sum()
    h = h.normalized(density=True)
    h.scatter(color="k", linewidth= 3, alpha= 1.)
    #ax.errorbar(bin_centers, histvals)
    ax.set_yscale("log")

    x_right_edge = max(h.bincenters[h.bincontent > 0])
    y_highest = max(h.bincontent)*10
    ax.set_xlim(xmax=x_right_edge,xmin=0)
    ax.grid(True)
    ax.set_ylim(ymin=1e-5,ymax=y_highest)
    sb.despine()
    ax.set_xlabel("Q [a. u.]")
    ax.set_ylabel("pdf")
    ax.text(8,1,"$\mu_{SPE}$ %4.2f" %params_duo[4])
    fig.savefig(sys.argv[2],bbox_inches="tight")
    p.show()



"""
Convenient operations

"""

import numpy as np
import scipy.integrate as integrate
from scipy.constants import elementary_charge as ECHARGE
from copy import deepcopy as copy

IMPEDANCE = 50


def average_wf(waveforms):
    """
    Get the average waveform

    Args:
        waveforms (list):

    Returns:
        np.ndarray
    """
    wf0 = copy(waveforms[0])
    for wf in waveforms[1:]:
        wf0 += wf

    return wf0 / float(len(waveforms))


def integrate_wf(header, waveform, method=integrate.simps, impedance = IMPEDANCE):
    """
    Integrate a waveform to get the total charge

    Args:
        header (dict):
        waveform (np.ndarray):

    Returns:
        float
    """
    integral = method(waveform, header["xs"], header["xincr"])
    return integral/impedance


def save_waveform(header, waveform, filename):
    """
    save a waveform together with its header

    Args:
        header (dict):
        waveform (np.ndarray):
        filename (str):
    Returns:
        None
    """
    np.save(filename, (header, waveform))
    return None


def load_waveform(filename):
    """
    load a waveform from a file

    Args:
        filenaame (str): 

    Returns:
        dict
    """
    head, wf = np.load(filename)
    return head, wf



"""
Prepare logging functionality for the module
"""

import sys
import logging
import os

LOGFORMAT = '[%(asctime)s] %(levelname)s: %(module)s(%(lineno)d):   %(message)s'


def get_logger(loglevel,logfile=None):
    """
    A root logger with a formatted output logging to stdout and a file

    Args:
        loglevel (int): 10,20,30,... the higher the less logging
        logfile (str): write logging to this file as well as stdout
    Returns:
        logging.logger
    """

    def exception_handler(exctype, value, tb):
        logger.critical("Uncaught exception", exc_info=(exctype, value, tb))

    logger = logging.getLogger(__name__)
    logger.setLevel(loglevel)
    ch = None
    for h in logger.handlers:
        if isinstance(h, logging.StreamHandler):
            ch = h
            break

    if ch is None:
        ch = logging.StreamHandler()

    ch.setLevel(loglevel)
    formatter = logging.Formatter(LOGFORMAT)
    ch.setFormatter(formatter)

    if logfile is not None:
        today = datetime.now()
        today = today.strftime("%Y-%m-%d_%H-%M")
        logend = ".log"
        if logfile.endswith(".log"):
            logfile.replace(".log",today+logend)
        else:
            logfile += (today + logend)
        logfilecount = 1

        # find a file name which does not exist yet
        while os.path.exists(logfile):
            logfile = logfile.replace("." + str(logfilecount -1),"")
            logfile = logfile +"." + str(logfilecount)
            logfilecount += 1
            if logfilecount >= 60:
                raise SystemError("More than 1 logfile per second, this is insane.. aborting")

        fh = logging.FileHandler(logfile)
        fh.setFormatter(formatter)
        fh.setLevel(loglevel)

    logger.addHandler(ch)
    if logfile is not None: logger.addHandler(fh)
    sys.excepthook = exception_handler
    logger.propagate = False
    return logger



"""
Package to read out TektronixDPO4104B oscilloscope 
"""

__version__ = "0.0.13"
__all__ = ["daq", "osci", "plotting", "tools", "fit"]

import appdirs as _appdirs
import shutil as _shutil
import os as _os

from matplotlib import get_configdir as mpl_configdir

from . import logging

_appdir = _os.path.split(__file__)[0]
_appname = _os.path.split(_appdir)[1]

# the config files
STYLE_BASEFILE_STD = _os.path.join(_appdir,"pyoscipresent.mplstyle")
STYLE_BASEFILE_PRS = _os.path.join(_appdir,"pyoscidefault.mplstyle")

logger = logging.get_logger(20)

def get_configdir():
    """
    Definges a configdir for this package under $HOME/.pyevsel
    """
    config_dir = _appdirs.user_config_dir(_appname)
    if not _os.path.exists(config_dir):
        _os.mkdir(config_dir)
    return config_dir

def install_styles(style_default=STYLE_BASEFILE_STD, \
                   style_present=STYLE_BASEFILE_PRS):
    """
    Sets up style files

    Keyword Args:
        style_default (str): location of style file to use by defautl
        style_present (str): location of style file used for presentations
        plots_config (str): configureation file for plots
        patternfile (str): location of patternfile with file patterns to search and read
    """

    logger.info("Installing styles...")
    cfgdir = get_configdir()
    mpl_styledir = _os.path.join(mpl_configdir(),"stylelib")
    for f in style_default, style_present:
        assert _os.path.exists(f), "STYLEFILE {} missing... indicates a problem with some paths or corrupt packege.\
                                    Check source code location".format(f)
        _shutil.copy(f,mpl_styledir)

install_styles()



from . import osci

class Agilent33220A(osci.AbstractBaseOscilloscope):

    def __init__(self, ip="10.25.21.168"):
        """
        An Agilent function generator

        Args:
            ip: The port where this instrument is listening on
        """
        osci.AbstractBaseOscilloscope.__init__(self, ip=ip)
        


"""
Use the scope as a DAQ

"""
from . import osci
from . import tools
from . import commands as cmd
from . import plotting
from . import logging as daqlog

logger = daqlog.get_logger(20)

import time
import re
import numpy as np
import pylab as p
try:
    import zmq
except ImportError:
    logger.warning("No zmq available!")


from socket import timeout as TimeoutError
from datetime import datetime


bar_available = False

try:
    import pyprind
    bar_available = True
except ImportError:
    logger.warning("No pyprind available")

try:
    from functools import reduce
except ImportError:
    logger.warning("Can not import functools, this might be python 2.7?")

# helpers





class Event(object):
    """
    DAQ will return events when triggered.
    """

    def __init__(self, use_datetime=False):
        """
        Keyword Args:
            use_datetime (bool): if True, give timestamp with datetime.datetime
        """
        self.use_datetime = use_datetime
        self.data = dict()
        self.timestamp = None

    def timestamp_it(self):
        """
        Give it a timestamp! Time in seconds

        Returns:
            None
        """
        if self.use_datetime:
            self.timestamp = datetime.now()
        else:
            self.timestamp = time.monotonic()


class DAQ(object):
    """
    A virtual DAQ using an oscilloscope
    """

    def __init__(self):
        """
        Initialize a new collector for instrument data
        """
        self.channels = dict()

    def register_instrument(self, instrument, label="instrument"):
        """
        Register an instrument and assign a channel to it. Instruments must have a pull()
        method which allows to pull data from them at a certain event.

        Args:
            instrument (ducktype): needs to be configured already and must have a pull() method
            channel_name (int): identify the instrument under this registered channel

        Returns:
            None
        """
        assert label not in self.channels.keys(),\
            "Instrument with label {} already registered! Chose a different label".format(label)

        self.channels[label] = instrument

    def acquire(self, *pullargs, **pullkwargs):
        """
        Go through the instrument list and trigger their pull methods to build an event

        Keyword Args:
            **pullkwargs (dict): will be passed on the individual pull methods

        Returns:
            pyosci.Event
        """
        event = Event()
        for key in self.channels.keys():
            event.data[key] = self.channels[key].pull(*pullargs, **pullkwargs)
        return event

    def acquire_n_events(self, n_events, trigger_hook=lambda x:x,\
                         trigger_hook_args=(None,), pull_args=(), pull_kwargs={}):
        """
        Continuous acquisition. Acquires n events. Yields events. Use trigger hook to define a
        function to decide when data is returned.

        Args:
            n_events (int): Number of events to acquire
            trigger_hook (callable): Trigger condition
            trigger_hook_args (tuple): Arguments for the trigger condition
        Yields:
            Event
        """
        if bar_available:
            bar = pyprind.ProgBar(n_events, track_time=True, title='Acquiring waveforms...')

        for __ in range(n_events):
            trigger_hook(*trigger_hook_args)
            if bar_available:
                bar.update()
            yield self.acquire(*pull_args, **pull_kwargs)










"""
Communicate with oscilloscope via vxi11 protocol over LAN network
"""
import abc
import time
import numpy as np
import pylab as p
import vxi11
from six import with_metaclass

from . import commands as cmd
from . import logging
from . import plotting
from . import tools

from copy import copy

bar_available = False

try:
    import pyprind
    bar_available = True
except ImportError:
    pass
    #logger.warning("No pyprind available")

try:
    # Python 2
    from itertools import izip
except ImportError:
    # Python 3
    izip = zip

from functools import reduce

# abbreviations
aarg = cmd.add_arg
q = cmd.query

TCmd = cmd.TektronixDPO4104BCommands
RSCmd = cmd.RhodeSchwarzRTO1044Commands

BIG_NUMBER = 1e25 # A number larger than the amount of captured samples

def setget(command):
    """
    Shortcut to construct property object to wrap getters and setters
    for a number of settings

    Args:
        command (str): The command being used to get/set. Get will be a query
        value (str): The value to set

    Returns:
        property object
    """
    return property(lambda self: self._send(q(command)),\
                    lambda self, value: self._set(aarg(command, value)))


class AbstractBaseOscilloscope(with_metaclass(abc.ABCMeta, object)):
    """
    A oscilloscope with a high sampling rate in the order of several
    gigasamples. Defines the scope API the DAQ reiles on
    """

    def __init__(self, ip="169.254.68.19", loglevel=20):
        """
        Connect to the scope via its socket server

        Args:
            ip (str): ip of the scope
        """
        self.ip = ip
        self.connect_trials = 0
        self.wf_buff_header = None # store a waveform header in case they are all the same
        self.instrument = vxi11.Instrument(ip)
        self.active_channel = None
        self.logger = logging.get_logger(loglevel)

    def reopen_socket(self):
        """
        Close and reopen the socket after a timeout

        Returns:
            None
        """
        self.logger.debug("Reopening socket on ip {}".format(self.ip))
        self.instrument = vxi11.Instrument(self.ip)

    def _send(self,command):
        """
        Send command to the scope. Raises socket.timeout error if
        it had failed too often

        Args:
            command (str): command to be sent to the
                           scope

        """
        if self.connect_trials == self.MAXTRIALS:
            self.connect_trials = 0
            raise vxi11.vxi11.Vxi11Exception("TimeOut")

        self.logger.debug("Sending {}".format(command))
        try:
            response = self.instrument.ask(command)
        except Exception as e:
            self.reopen_socket()
            response = self.instrument.ask(command)
            self.connect_trials += 1

        return response

    def _set(self, command):
        """
        Send a command bur return no response

        Args:
            command (str): command to be send to the scope

        Returns:
            None
        """
        self.logger.debug("Sending {}".format(command))
        self.instrument.write(command)

    def ping(self):
        """
        Check if oscilloscope is connected
        """

        ping = self._send(cmd.WHOAMI)
        self.logger.info("Scope responds to {} with {}".format(cmd.WHOAMI, ping))
        return True if ping else False

    def __repr__(self):
        """
        String representation of the scope
        """
        ping = self._send(cmd.WHOAMI)
        return "<" + ping + ">"

    @abc.abstractmethod
    def select_channel(self, channel):
        """
        Select a channel for the data acquisition

        Args:
            channel (int): Channel number

        Returns:
            None
        """
        return

    @abc.abstractproperty
    def samplingrate(self):
        """
        Get the current sampling rate

        Returns:
            float (GSamples/sec)
        """
        return

    def __del__(self):
        """
        Destructor, close connection explicitely.

        Returns:
            None
        """
        self.instrument.close()


class Waveform(object):
    """
    A non-oscilloscope dependent representation of a measured
    waveform
    """
    header = dict()

    def __init__(self, header, curvedata):
        """
        Args:
            header (dict): Metadata, like xs, units, etc.
            curvedata (np.ndarray): The voltage data
        """

        self.header = header
        self.data = curvedata

    @property
    def ns(self):
        """
        Return the signal recording time with ns precision

        Returns:
            np.ndarray
        """
        return

    @property
    def mV(self):
        """
        Return the recorded voltages per bin

        Returns:
            np.ndarray
        """
        return

    @property
    def bins(self):
        """
        Return digitizer time bin numbers (arbitrary scale)

        Returns:
            np.ndarray
        """
        return


class TektronixDPO4104B(AbstractBaseOscilloscope):
    """
    Oscilloscope of type DPO4104B manufactured by Tektronix
    """

    # constants used by the socket connection
    MAXTRIALS = 5


    # setget properties
    source = setget(cmd.SOURCE)
    data_start = setget(cmd.DATA_START)
    data_stop = setget(cmd.DATA_STOP)
    waveform_enc = setget(cmd.WF_ENC)
    #fast_acquisition = setget(cmd.ACQUIRE_FAST_STATE)
    acquire = setget(cmd.RUN)
    acquire_mode = setget(TCmd.ACQUISITON_MODE)
    data = setget(cmd.DATA)
    trigger_frequency_enabled = setget(TCmd.TRIGGER_FREQUENCY_ENABLED)
    histbox = setget(cmd.HISTBOX)
    histstart = setget(cmd.HISTSTART)
    histend = setget(cmd.HISTEND)

    def __init__(self, ip, loglevel=20):
        AbstractBaseOscilloscope.__init__(self, ip, loglevel=loglevel)
        self.active_channel = TCmd.CH1
        self._header_buff = False
        self._wf_buff = np.zeros(len(self.waveform_bins))
        self._data_start_stop_buffer = (None, None)

        # FIXME: future extension
        self._is_running = False
        self._acquisition_single = False

        # fill the buffer
        self.fill_header_buffer()
        self.fill_buffer()

    def fill_header_buffer(self):
        self._header_buff = self.wf_header()

    def trigger_single(self):
        self.acquire_mode = cmd.RUN_SINGLE

    def trigger_continuous(self):
        self.acquire_mode = cmd.RUN_CONTINOUS
        self.acquire = cmd.START_ACQUISITIONS

    def _trigger_acquire(self):
        """
        Acquire one single waveform

        Returns:
            None
        """
        self.acquire = "ON"

    @staticmethod
    def _parse_wf_header(header):
        """
        Parse a waveform header send by our custom WF_HEADER command
        The reason why we are not using WFM:Outpre is that the documentation
        was not so sure about how its response might look

        Args:
            head (str): the result of a WF_HEADER command

        Returns:
            dict
        """
        head = header.split(";")
        keys = ["bytno", "enc", "npoints", "xzero", "xincr", "yzero", "yoff", \
                "ymult", "xunit", "yunit"]

        assert len(head) == len(keys), "Cannot read out all the header info I want!"

        parsed = dict(zip(keys, head))
        for k in parsed:
            try:
                f = float(parsed[k])
                parsed[k] = f
            except ValueError:
                continue

            # get rid of extra " in units
            parsed["xunit"] = parsed["xunit"].replace('"', '')
            parsed["yunit"] = parsed["yunit"].replace('"', '')

            # also some are ints
            parsed["npoints"] = int(parsed["npoints"])
        return parsed

    def select_channel(self, channel):
        """
        Select the channel for the readout

        Args:
            channel (int): Channel number (1-4)

        Returns:
            None
        """

        assert 0 < channel < 5, "Channel value has to be 1-4"

        channel_dict = {1 : TCmd.CH1, 2: TCmd.CH2, 3: TCmd.CH3, 4: TCmd.CH4}
        self.source = channel_dict[channel]
        self.active_channel = channel_dict[channel]
        self.logger.info("Selecting channel {}".format(self.source))
        return None

    def _select_active_channel(self):
        """
        Pick the channel which is intended to be used

        Returns:

        """
        self.source = self.active_channel

    @property
    def samplingrate(self):
        """
        The samplingrate in GSamples/S

        Returns:
            float
        """
        head = self.wf_header()
        self.logger.debug("Got samplingrate of {}".format(1./head["xincr"]))
        return 1./head["xincr"]

    @property
    def trigggerrate(self):
        """
        The rate the scope is triggering. The scope in principle provides this number,
        however we have to work around it as it does not work reliably

        Keyword Arguments:
            interval (int): time interval to integrate measurement over in seconds

        Returns:
            float
        """
        self.logger.debug("The returned value is instantanious!\n "
                            "For serious measurements, gather some statistics!")
        self.trigger_frequency_enabled = TCmd.ON
        freq = float(self._send(TCmd.TRIGGER_FREQUENCYQ))
        return freq

    def reset_acquisition_window(self):
        """
        Reset the acquisition window to the maximum possible acquisition window
        Returns:In
            None
        """
        self.data_start = 0
        self.data_stop = BIG_NUMBER # temporarily set this to a big bogus number
                                    # this will result in the correct value for
                                    # "npoints" later
        head = self.wf_header()
        self.set_acquisition_window(0, head["npoints"])

    @property
    def time_binwidth(self):
        """
        Get the binwidth of the time - that is sampling rate

        Returns:
            float
        """
        head = self.wf_header()
        return float(head["xincr"])

    @property
    def waveform_bins(self):
        """
        Get the time bin numbers for the waveform voltage data

        Returns:
            np.ndarray
        """
        head = self.wf_header()
        bin_zero = self.data_start
        bins = np.linspace(int(self.data_start), int(self.data_stop), int(head["npoints"]))
        return bins

    @property
    def waveform_times(self):
        """
        Get the time for the waveform bins

        Returns:
            np.ndarray
        """
        head = self.wf_header()
        return head["xs"]

    @property
    def histogram(self):
        """
        Return a histogram which might be recorded
        by the scope
        """
        start = self.histstart
        end = self.histend
        bincontent = self._send(cmd.HISTDATA)
        assert None not in [start,end,bincontent],\
                   "Try again! might just be a hickup {} {} {}".format(start,end,bincontent)

        bincontent = np.array([int(b) for b in bincontent.split(",")])#
        start = float(start)
        end = float(end)    
        nbins = len(bincontent)
        if start > end:
            print ("Swapping start and end...")
            tmpstart = copy(start)
            start = end
            end = tmpstart
            del tmpstart

        print("Found histogram with {} entries from {:4.2e} to {:4.2e}".format(nbins,start, end))

        l_binedges = np.linspace(start,end,nbins + 1)[:-1]
        r_binedges = np.linspace(start,end,nbins + 1)[1:]
        bincenters = r_binedges + (r_binedges - l_binedges)/2.
        return bincenters, bincontent 

    def wf_header(self, absolute_timing=False):
        """
        Get some meta information about the *next incoming wavefrm*

        Keyword Args:
            absolute_timing (bool): header["xs"] starts with header["xzero"], 0 otherwiese

        Returns:
            dict
        """
        header = self._send(TCmd.WF_HEADER)
        header = self._parse_wf_header(header)
        self.wf_buff_header = header
        #header["xs"] = np.ones(len(header["npoints"]))*header["xzero"]
        if absolute_timing:
            xs = np.ones(header["npoints"])*header["xzero"]
        else:
            # relative timing?
            xs = np.zeros(int(header["npoints"]))
        # FIXME: There must be a better way
        for i in range(int(header["npoints"])):
            xs[i] += i*header["xincr"]

        header["xs"] = xs
        return header

    def acquire_waveform(self, buff_header=False):
        """
        Get the waveform data

        Keyword Args:
            buff_header: buffer the header (do NOT query the header every time)
                         WARNING: This is only correct if there are no changes
                                  to the way the acquisition is made

        Returns:
            np.ndarray
        """

        waveform = self._send(cmd.CURVE)
        waveform = np.array([float(k) for k in waveform.split(",")])
        if buff_header:
            header = self._header_buff
            assert header is not None, "Nothing in header buffer, call fill_header_buffer"
        else:
            header = self.wf_header()

        # from the docs
        # Value in YUNit units = ((curve_in_dl - YOFf) * YMUlt) + YZEro
        waveform = (waveform - (np.ones(len(waveform))*header["yoff"]))\
                   *(np.ones(len(waveform))*header["ymult"])\
                   +(np.ones(len(waveform))*header["yzero"])

        return waveform

    def set_acquisition_window(self, start, stop):
        """
        Set the acquisition window in bin number

        Args:
            start (int): start bin
            stop (int): stop bin

        Returns:
            None
        """
        self.data_start = start
        self.data_stop = stop
        self._wf_buff = np.zeros(len(self.waveform_bins))
        self._data_start_stop_buffer = (start, stop)
        self.logger.info("Set acquisition window to {} - {}".format(self.data_start, self.data_stop))

    def set_acquisition_window_from_internal_buffer(self):
        """
        Use the internal buffer to set the data acquisition window. Might be necessary
        if the channel was switched in the meantime

        Returns:
            None
        """
        self.data_start, self.data_stop = self._data_start_stop_buffer
        #self.fill_header_buffer()
        #self.fill_bu

    def set_feature_acquisition_window(self, leading, trailing, n_waveforms=20):
        """
        Set the acquisition window around the most prominent feature in the waveform

        Args:
            leading (float): leading ns before the most prominent feature
            trailing (float): trailing ns after the most prominent feature

        Keyword Args
            n_waveforms (int): average over n_waveforms to identify the most prominent feature

        Returns:
            None
        """
        self.reset_acquisition_window()
        xs, avg = self.average_waveform(n=n_waveforms)
        wf_bins = self.waveform_bins

        abs_avg = abs(avg)
        feature_y = max(abs_avg)
        #feature_x = xs[abs_avg == feature_y]
        feature_x_bin = wf_bins[abs_avg == feature_y]
        bin_width = self.time_binwidth
        leading_bins = 1e-9*float(leading)/bin_width
        trailing_bins = 1e-9*float(trailing)/bin_width
        data_start = int(feature_x_bin - leading_bins)
        data_stop = int(feature_x_bin + trailing_bins)
        self.set_acquisition_window(data_start, data_stop)
        return None

    def make_n_acquisitions(self, n,\
                            trials=20, return_only_charge=False,\
                            single_acquisition=True):
        """
        Acquire n waveforms

        Args:
            n (int): Number of waveforms to acquire

        Keyword Args:
            trials (int): Set breaking condition when to abort acquisition
            return_only_charge (bool): don't get the wf, but only integrated charge instead
            single_acquisition (bool): use the scopes single acquisition mode

        Returns:
            list: [wf_1,wf_2,...]

        """
        wforms = list()
        acquired = 0
        trial = 0
        if bar_available:
            bar = pyprind.ProgBar(n, track_time=True, title='Acquiring waveforms...')
        if single_acquisition:
            self.trigger_single()
        else:
            self.trigger_continuous()
        wf_buff = 0
        while acquired < n:
            try:
                if single_acquisition:
                    self.acquire = "ON"
                wf = self.acquire_waveform(buff_header=True)
                if (wf[0]*np.ones(len(wf)) - wf).sum() == 0:
                    continue # flatline test
                if (wf - wf_buff).sum() == 0:
                    continue # test if scope just returned the
                             # same waveform again
                if return_only_charge:
                     wf = tools.integrate_wf(header, wf)
                wf_buff = wf
                wforms.append(wf)
                acquired += 1
                if bar_available:
                    bar.update()

            except Exception as e:
                self.logger.critical("Can not acquire wf..{}".format(e))
                trial += 1
            if trial == trials:
                break
        if bar_available:
            print(bar)
        return wforms

    def average_waveform(self,n=10):
        """
        Acquire some waveforms and take the average

        Keyword Args.
            n (int): number of waveforms to average over

        Returns:
            tuple(np.array). xs, ys

        """

        wf = self.make_n_acquisitions(n, single_acquisition=False)
        xs = self.waveform_times
        len_wf = [len(w) for w in wf]
        wf = [w for w in wf if len(w) == min(len_wf)]
        avg = reduce(lambda x, y: x+y, wf)/n
        return xs, avg

    def show_waveforms(self,n=5):
        """
        Demonstration function: Will use pylab show to
        plot some acquired waveforms

        Keyword Args:
            n (int): number of waveforms to show

        Returns:
            None
        """

        wf = self.make_n_acquisitions(n, single_acquisition=False)
        head = self.wf_header()
        for i in range(len(wf)):
            try:
                plotting.plot_waveform(head, wf[i])
            except Exception as e:
                print(e)

        p.show()

    def fill_buffer(self):
        """


        Returns:

        """
        self._wf_buff = self.acquire_waveform()


    def pull(self, buff_header=True, use_buffered_acq_window=True,
             use_channel_info=True):
        """
        Fit in the API for the DAQ. Returns waveform data

        Keyword Args:
            buff_header (bool): buffer the header for subsequent acquisition without
                                changing the parameters of the acquistion (much faster)
            FIXME! Default value of this should be False, however requires DAQ API change
            use_buffered_acq_window (bool): set this flag to cache the length of the acquisition window
                                            internally so that it does not get resetted when switching channels
            use_channel_info (bool): select the channel on each submit
        Returns:
            dict
        """
        user_error_msg = """ This pull method is designed to be used in single acquisition mode."""

        assert self.acquire_mode == cmd.RUN_SINGLE, user_error_msg

        # FIXME: The buffer mechanism fails if this is the first
        # waveform at all.
        data = dict()
        if use_channel_info:
            self._select_active_channel()
        if use_buffered_acq_window:
            self.set_acquisition_window_from_internal_buffer()
        #while True:
        wf = self.acquire_waveform(buff_header=buff_header)

            #if (wf[0]*np.ones(len(wf)) - wf).sum() == 0:
            #    continue # flatline test
            #elif (wf - self._wf_buff).sum() == 0:
            #    self._wf_buff = wf
            #    continue # test if scope just returned the
            #             # same waveform again
            #else:
            #    self._wf_buff = wf
            #    break
        if buff_header:
            data.update(self._header_buff)
        else:
            data.update(self.wf_header())
        data["waveform"] = wf
        return data


class RhodeSchwarzRTO1044(AbstractBaseOscilloscope):
    """
    Made by Rhode&Schwarz, scope with sampling rate up to 20GSamples/s
    """

    def __init__(self, ip):
        AbstractBaseOscilloscope.__init__(self,ip)
        self.active_channel = RSCmd.CH1

    def select_channel(self, channel):
        """
        Select the channel for the readout

        Args:
            channel (int): Channel number (1-4)

        Returns:
            None
        """
        channel_dict = {1: RSCmd.CH1, 2: RSCmd.CH2, 3: RSCmd.CH3, 4: RSCmd.CH4}
        self.active_channel = channel_dict[channel]

    def get_waveform(self):
        """
        Get the voltage values for a single waveform

        Returns:
            np.ndarray
        """
        wf_commnad = aarg(self.active_channel,RSCmd.CURVE)
        raw_wf = self._send(wf_command)
        pairs = izip(*[iter(raw_wf.split(","))]*2)
        times, volts = [],[]
        for val in pairs:
            volts.append(float(val[1]))
        return np.array(volts)



from setuptools import setup

from pyosci import __version__

setup(name='pyosci',
      version=__version__,
      description='Readout a TektronixDPO4104B oscilloscope',
      long_description='Use the oscilloscope for readout of waveforms bascially as a daq. Provides an easy to extend API to inlcude more functionality',
      author='Achim Stoessl',
      author_email="achim.stoessl@gmail.com",
      url='https://github.com/achim1/pyosci',
      #download_url="pip install pyosci",
      install_requires=['numpy>=1.11.0',
                        'matplotlib>=1.5.0',
                        'appdirs>=1.4.0',
                        'pyprind>=2.9.6',
                        'six>=1.1.0'],
      license="GPL",
      platforms=["Ubuntu 14.04","Ubuntu 16.04"],
      classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering :: Physics"
              ],
      keywords=["oscilloscope", "daq",\
                "TektronixDPO4104B", "Tektronix",\
                "readout", "physics", "engineering", "SCPI", "VISA"],
      packages=['pyosci'],
      #scripts=[],
      package_data={'pyosci': ['pyoscidefault.mplstyle','pyoscipresent.mplstyle']}
      )



"""
Convenient plot functions

"""

import pylab as p
import seaborn.apionly as sb

from copy import copy

p.style.use("pyoscidefault")


def plot_waveform(wf_header, wf_data,\
                  fig=None,savename=None,\
                  use_mv_and_ns=True,color=sb.color_palette("dark")[0]):
    """
    Make a plot of a single acquisition

    Args:
        wf_header (dict): custom waveform header
        wf_data (np.ndarray): waveform data

    Keyword Args:
        fig (pylab.figure): A figure instance
        savename (str): where to save the figure (full path)
        use_mv_and_ns (bool): use mV and ns instead of V and s
    Returns:
        pylab.fig
    """

    if fig is None:
        fig = p.figure()
    ax = fig.gca()

    # if remove_empty_bins:
    #    bmin = min(bincenters[bincontent > 0])
    #    bmax = max(bincenters[bincontent > 0])
    #    bincenters = bincenters[np.logical_and(bincenters >= bmin, bincenters <= bmax)]
    #    bincontent = bincontent[np.logical_and(bincenters >= bmin, bincenters <= bmax)]

    xlabel = wf_header["xunit"]
    ylabel = wf_header["yunit"]
    xs = copy(wf_header["xs"])
    ys = copy(wf_data)

    if xlabel == "s" and ylabel == "V" and use_mv_and_ns:
        xs *= 1e9
        ys *= 1e3
        xlabel = "ns"
        ylabel = "mV"
    ax.plot(xs, ys, color=color)
    ax.grid()
    sb.despine(fig)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    p.tight_layout()
    if savename is not None:
        fig.savefig(savename)
    return fig

#####################################################

def plot_histogram(bincenters,bincontent,\
                   fig=None,savename="test.png",\
                   remove_empty_bins=True):
    """
    Plot a histogram returned by TektronixDPO4104B.get_histogram
    Use pylab.plot

    Args:
        bincenters (np.ndarray); bincenters (x)
        bincontent (np.ndarray): bincontent (y)

    Keyword Args:
        fig (pylab.figure): A figure instance
        savename (str): where to save the figure (full path)
        remove_empty_bins (bool): Cut away preceeding and trailing zero bins


    """


    if fig is None:
        fig = p.figure()
    ax = fig.gca()
    if remove_empty_bins:
        bmin = min(bincenters[bincontent > 0])
        bmax = max(bincenters[bincontent > 0])
        bincenters = bincenters[np.logical_and(bincenters >= bmin, bincenters <= bmax)]
        bincontent = bincontent[np.logical_and(bincenters >= bmin, bincenters <= bmax)]

    ax.plot(bincenters,bincontent,color=sb.color_palette("dark")[0])
    ax.grid()
    sb.despine(fig)
    ax.set_xlabel("amplitude")
    ax.set_ylabel("log nevents ")
    p.tight_layout()
    fig.savefig(savename)
    return fig





"""
A namespace for oscilloscope string commands. The commands are send as ASCII
to the scope using a socket connection
"""

query = lambda cmd : cmd  + "?"

def add_arg(cmd, arg):
    """
    Add an argument to a command string. Concat the two.

    Args:
        cmd (str): The base command
        arg (str): An argument

    Returns:
        str
    """
    if not isinstance(arg,str):
        arg = str(arg)

    if len(cmd.split()) != 1:
        cmd = cmd + "," + arg
    else:
        cmd = cmd + " " + arg
    return cmd


def concat(*cmd):
    """
    Combine several commands

    Args:
        cmd: list of commands

    Returns:
        str
    """
    cmd = list(cmd)
    cmd[0].replace(":","")
    concated = cmd[0] + ";"
    cmd.remove(cmd[0])
    for c in cmd:
        concated += ":" + str(c) + "; "
    return concated


#def decode(response):
#    """
#    Decode osciloscope response from bytes to a string, cleaning
#    trailing EOL characters.
#
#    Args:
#        response (bytes): recieved from oscilloscope
#
#    Returns:
#        str
#    """
#    if response is None:
#        return response
#    if isinstance(response,bytes):
#        response = response.decode("utf8")
#    response = clean_response(response)
#    if response in [False, "\r\n", "\n\r"]:
#        response = None
#    return response


def clean_response(response):
    """
    Remove some EOL remainders from the resulting scope response

    Args:
        response (str): response from the scope

    Returns:
        str
    """

    response = response.replace("\r\n>", "")
    response = response.replace("\r\n", "")
    response = response.replace("\n\r", "")
    response = response.replace(">", "")
    response = response.replace("<", "")
    response = response.replace("\n", "")
    return response


def parse_curve_data(header, curve):
    """
    Make sense out of that what is returned by CURVE. This works only
    if the scope is set to return the data in ASCII, not binary.

    Args:
        header (dict): a parsed header
        curv (str): returned by CURVE?

    Returns:
        tuple: np.ndarray xs, np.ndarray values
    """

    # Value in YUNit units = ((curve_in_dl - YOFf) * YMUlt) + YZEro

    curve.replace("#", "")


#def encode(cmd):
#    """
#    Append trailing return carriage and convert
#    to bytes. Trailing EOL characters are important
#    otherwise the scope will listen for input commands
#    endlessly.
#
#    Args:
#        cmd (str): command to be sanitized
#
#    Returns:
#        bytearray
#    """
#
#    if not cmd.endswith("\r\n"):
#        cmd = cmd + "\r\n"
#
#    # python3 needs byterepresentation for socket
#    return cmd.encode("utf8")

def histbox_coordinates(left, top, right, bottom):
    """
    Create a string for the box coordinates for the
    histogram set up by the scope itself. The result
    can be send to the scope to set the box.

    Args:
        left (int):
        top (int):
        right (int):
        bottom (int):

    Returns:
        str
    """
    command = concat([left, top, right, bottom])
    return command



#
# COMMANDS!
# (add more here if needed. See the programming manual)
# Queries should end with "Q"
#

WHOAMI = "*IDN?"
SOURCE = "DATa:SOUrce"
DATA_START = "DATa:STARt"
DATA_STOP = "DATa:STOP"
DATA_STARTQ = query(DATA_START)
DATA_STOPQ = query(DATA_STOP)
SOURCEQ = query(SOURCE)
HISTSTART = "HIStogram:STARt"
HISTEND = "HIStogram:END"
HISTDATA = "HIStogram:DATa?"
WAVEFORM = "WAVFrm?"
CURVE = "CURVe?"
WF_OUTPREQ = "WFMOutre?"
WF_XINCRQ = "WFMOutpre:XINcr?"
WF_XUNITQ = "WFMOutpre:XUNit?"
WF_XZEROQ = "WFMOutpre:XZEro?"
WF_YOFFQ = "WFMOutpre:YOFf?"
WF_YZEROQ = "WFMOutpre:YZEro?"
WF_YMULTQ = "WFMOutpre:YMUlt?"
WF_YUNITQ = "WFMOutpre:YUNit?"
WF_ENC = "WFMOutpre:ENCdg"
WF_ENCQ = query(WF_ENC)
WF_NPOINTS = "WFMOutpre:NR_Pt"
WF_NPOINTSQ = query(WF_NPOINTS)
ACQUIRE = "ACQuire"
ACQUIRE_FAST_STATE = "ACQuire:FASTAcq:STATE"
ACQUIRE_MAX_SAMPLERATEQ = "ACQuire:MAXSamplerate?"
RUN = "ACQuire:STATE"
WF_BYTQ = "WFMOutpre:BYT_Nr?" # 1 or 2 for single or double prec
ACQUIRE_STOP = "ACQuire:STOPAfter"
HISTBOX = "HIStogram:BOX"
TRG_RATEQ = "TRIGger:FREQuency?"
# command arguments

RUN_CONTINOUS = "RUNSTop"
RUN_SINGLE = "SEQ" # sequence
START_ACQUISITIONS = "RUN"
STOP_ACQUISITIONS = "STOP"
N_ACQUISITIONS = "ACQuire:NUMACq?"
ACQUIRE_ONE = "ON"
SNAP = "SNAP"
DATA = "DATA"
OFF = "0"
ON = "1"



ASCII = "ASCii"
BINARY = "BINary"
SINGLE_ACQUIRE = "1"

# combined commands


class TektronixDPO4104BCommands(object):
    """
    Namespace for the commands for the TektronixDP04104B
    """
    WF_HEADER = concat(WF_BYTQ, WF_ENCQ, WF_NPOINTSQ, WF_XZEROQ, WF_XINCRQ, \
                       WF_YZEROQ, WF_YOFFQ, WF_YMULTQ, \
                       WF_XUNITQ, WF_YUNITQ)
    CH1 = "CH1"
    CH2 = "CH2"
    CH3 = "CH3"
    CH4 = "CH4"
    ON = "ON"
    OFF = "OFF"
    ONE = "1"
    ZERO = "0"
    TRIGGER_FREQUENCY_ENABLED = "DISplay:TRIGFrequency"
    TRIGGER_FREQUENCYQ = "TRIGger:FREQuency?"
    ACQUISITON_MODE = "ACQuire:STOPAfter"

class RhodeSchwarzRTO1044Commands(object):
    """
    Namespace for the commands for the RhodeSchwarz oscilloscope
    """
    CH1 = "CHAN1"
    CH2 = "CHAN2"
    CH3 = "CHAN3"
    CH4 = "CHAN4"
    WAVEFORM = "DATA?"
    WF_HEADER = "DATA:HEADer?"
    CURVE = "DATA:VALues?"

class KeysightE3631APowerSupplyCommands(object):
    """
    Namespace for the commands of the KeysightE3631APowerSupply
    """
    ERROR_STATEQ = "SYST:ERR?"
    OUTPUT = "OUTPUT:STATE"
    ON = "ON"
    OFF = "OFF"
    P6 = "P6V" # the 6V channel
    P25 = "P25V" # the 25V channel
    N25 = "N25V" # the 25V channel but with negative polarity
    APPLY = "APPLY"
    CHANNEL = "INST"
    VOLT = "VOLT"
    MEASURE = "MEASURE"
    CURRENT = "CURRENT"
    DC = "DC"